@@ -94,7 +94,11 @@ enum
   PROP_PROXY_IP,
   PROP_PROXY_PORT,
   PROP_PROXY_USERNAME,
-  PROP_PROXY_PASSWORD
+  PROP_PROXY_PASSWORD,
+#ifdef HAVE_GUPNP
+  PROP_UPNP,
+  PROP_UPNP_TIMEOUT
+#endif
 };
 
 
@@ -111,11 +115,28 @@ enum
 
 static guint signals[N_SIGNALS];
 
+static GStaticRecMutex agent_mutex = G_STATIC_REC_MUTEX_INIT;    /* Mutex used for thread-safe lib */
+
 static gboolean priv_attach_stream_component (NiceAgent *agent,
     Stream *stream,
     Component *component);
 static void priv_detach_stream_component (Stream *stream, Component *component);
 
+static void priv_free_upnp (NiceAgent *agent);
+
+
+void agent_lock (void)
+{
+  g_static_rec_mutex_lock (&agent_mutex);
+}
+
+void agent_unlock (void)
+{
+  g_static_rec_mutex_unlock (&agent_mutex);
+}
+
+
+
 StunUsageIceCompatibility
 agent_to_ice_compatibility (NiceAgent *agent)
 {
@@ -292,7 +313,7 @@ nice_agent_class_init (NiceAgentClass *klass)
         "stun-pacing-timer",
         "STUN pacing timer",
         "Timer 'Ta' (msecs) used in the IETF ICE specification for pacing candidate gathering and sending of connectivity checks",
-        1, 0xffffffff, 
+        1, 0xffffffff,
 	NICE_AGENT_TIMER_TA_DEFAULT,
         G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
@@ -302,7 +323,7 @@ nice_agent_class_init (NiceAgentClass *klass)
         "max-connectivity-checks",
         "Maximum number of connectivity checks",
         "Upper limit for the total number of connectivity checks performed",
-        0, 0xffffffff, 
+        0, 0xffffffff,
 	0, /* default set in init */
         G_PARAM_READWRITE));
 
@@ -348,6 +369,27 @@ nice_agent_class_init (NiceAgentClass *klass)
         NULL,
         G_PARAM_READWRITE));
 
+#ifdef HAVE_GUPNP
+   g_object_class_install_property (gobject_class, PROP_UPNP,
+      g_param_spec_boolean (
+        "upnp",
+        "Use UPnP",
+        "Whether the agent should use UPnP to open a port in the router and "
+        "get the external IP",
+	TRUE, /* enable UPnP by default */
+        G_PARAM_READWRITE| G_PARAM_CONSTRUCT));
+
+  g_object_class_install_property (gobject_class, PROP_UPNP_TIMEOUT,
+      g_param_spec_uint (
+        "upnp-timeout",
+        "Timeout for UPnP discovery",
+        "The maximum amount of time to wait for UPnP discovery to finish before "
+        "signaling the candidate-gathering-done signal",
+        100, 60000,
+	200,
+        G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+#endif
+
   /* install signals */
 
   /**
@@ -510,6 +552,7 @@ nice_agent_init (NiceAgent *agent)
   agent->conncheck_timer_source = NULL;
   agent->keepalive_timer_source = NULL;
   agent->refresh_list = NULL;
+  agent->media_after_tick = FALSE;
 
   agent->compatibility = NICE_COMPATIBILITY_DRAFT19;
 
@@ -520,8 +563,6 @@ nice_agent_init (NiceAgent *agent)
 
   agent->rng = nice_rng_new ();
   priv_generate_tie_breaker (agent);
-
-  g_static_rec_mutex_init (&agent->mutex);
 }
 
 
@@ -546,7 +587,7 @@ nice_agent_get_property (
 {
   NiceAgent *agent = NICE_AGENT (object);
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   switch (property_id)
     {
@@ -603,11 +644,21 @@ nice_agent_get_property (
       g_value_set_string (value, agent->proxy_password);
       break;
 
+#ifdef HAVE_GUPNP
+    case PROP_UPNP:
+      g_value_set_boolean (value, agent->upnp_enabled);
+      break;
+
+    case PROP_UPNP_TIMEOUT:
+      g_value_set_uint (value, agent->upnp_timeout);
+      break;
+#endif
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
 
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
 }
 
 
@@ -620,7 +671,7 @@ nice_agent_set_property (
 {
   NiceAgent *agent = NICE_AGENT (object);
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   switch (property_id)
     {
@@ -698,11 +749,20 @@ nice_agent_set_property (
       agent->proxy_password = g_value_dup_string (value);
       break;
 
+#ifdef HAVE_GUPNP
+    case PROP_UPNP_TIMEOUT:
+      agent->upnp_timeout = g_value_get_uint (value);
+      break;
+
+    case PROP_UPNP:
+      agent->upnp_enabled = g_value_get_boolean (value);
+      break;
+#endif
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
 
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
 
 }
 
@@ -744,7 +804,14 @@ void agent_gathering_done (NiceAgent *agent)
     }
   }
 
+#ifdef HAVE_GUPNP
+  if (agent->discovery_timer_source == NULL &&
+      agent->upnp_timer_source == NULL) {
+    agent_signal_gathering_done (agent);
+  }
+#else
   agent_signal_gathering_done (agent);
+#endif
 }
 
 void agent_signal_gathering_done (NiceAgent *agent)
@@ -860,7 +927,8 @@ priv_add_new_candidate_discovery_stun (NiceAgent *agent,
       stun_agent_init (&cdisco->stun_agent, STUN_ALL_KNOWN_ATTRIBUTES,
           STUN_COMPATIBILITY_RFC3489, 0);
 
-      nice_debug ("Agent %p : Adding new srv-rflx candidate discovery %p\n", agent, cdisco);
+      nice_debug ("Agent %p : Adding new srv-rflx candidate discovery %p\n",
+          agent, cdisco);
       agent->discovery_list = modified_list;
       ++agent->discovery_unsched_items;
     }
@@ -931,7 +999,6 @@ priv_add_new_candidate_discovery_turn (NiceAgent *agent,
               socket = nice_http_socket_new (socket, &turn->server,
                   agent->proxy_username, agent->proxy_password);
             } else {
-              /* TODO add HTTP support */
               nice_socket_free (socket);
               socket = NULL;
             }
@@ -1007,7 +1074,7 @@ nice_agent_add_stream (
   GSList *modified_list = NULL;
   guint ret = 0;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   stream = stream_new (n_components);
   if (stream) {
@@ -1026,7 +1093,7 @@ nice_agent_add_stream (
 
   ret = stream->id;
 
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1047,7 +1114,7 @@ nice_agent_set_relay_info(NiceAgent *agent,
   g_return_val_if_fail (password, FALSE);
   g_return_val_if_fail (type <= NICE_PROXY_TYPE_LAST, FALSE);
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   if (agent_find_component (agent, stream_id, component_id, NULL, &component)) {
     TurnServer *turn = g_slice_new0 (TurnServer);
@@ -1057,7 +1124,7 @@ nice_agent_set_relay_info(NiceAgent *agent,
       nice_address_set_port (&turn->server, server_port);
     } else {
       g_slice_free (TurnServer, turn);
-      g_static_rec_mutex_unlock (&agent->mutex);
+      agent_unlock();
       return FALSE;
     }
 
@@ -1072,10 +1139,148 @@ nice_agent_set_relay_info(NiceAgent *agent,
     component->turn_servers = g_list_append (component->turn_servers, turn);
   }
 
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return TRUE;
 }
 
+#ifdef HAVE_GUPNP
+
+static gboolean priv_upnp_timeout_cb (gpointer user_data)
+{
+  NiceAgent *agent = (NiceAgent*)user_data;
+  GSList *i;
+
+  agent_lock();
+
+  if (g_source_is_destroyed (g_main_current_source ())) {
+    agent_unlock ();
+    return FALSE;
+  }
+
+  nice_debug ("Agent %p : UPnP port mapping timed out", agent);
+
+  for (i = agent->upnp_mapping; i; i = i->next) {
+    NiceAddress *a = i->data;
+    nice_address_free (a);
+  }
+  g_slist_free (agent->upnp_mapping);
+  agent->upnp_mapping = NULL;
+
+  if (agent->upnp_timer_source != NULL) {
+    g_source_destroy (agent->upnp_timer_source);
+    g_source_unref (agent->upnp_timer_source);
+    agent->upnp_timer_source = NULL;
+  }
+
+  agent_gathering_done (agent);
+
+  agent_unlock();
+  return FALSE;
+}
+
+static void _upnp_mapped_external_port (GUPnPSimpleIgd *self, gchar *proto,
+    gchar *external_ip, gchar *replaces_external_ip, guint external_port,
+    gchar *local_ip, guint local_port, gchar *description, gpointer user_data)
+{
+  NiceAgent *agent = (NiceAgent*)user_data;
+  NiceAddress localaddr;
+  NiceAddress externaddr;
+
+  GSList *i, *j, *k;
+
+  agent_lock();
+
+  nice_debug ("Agent %p : Sucessfully mapped %s:%d to %s:%d", agent, local_ip,
+      local_port, external_ip, external_port);
+
+  if (!nice_address_set_from_string (&localaddr, local_ip))
+    goto end;
+  nice_address_set_port (&localaddr, local_port);
+
+  for (i = agent->upnp_mapping; i; i = i->next) {
+    NiceAddress *addr = i->data;
+    if (nice_address_equal (&localaddr, addr)) {
+      agent->upnp_mapping = g_slist_remove (agent->upnp_mapping, addr);
+      nice_address_free (addr);
+      break;
+    }
+  }
+
+  if (!nice_address_set_from_string (&externaddr, external_ip))
+    goto end;
+  nice_address_set_port (&externaddr, external_port);
+
+  for (i = agent->streams; i; i = i->next) {
+    Stream *stream = i->data;
+    for (j = stream->components; j; j = j->next) {
+      Component *component = j->data;
+      for (k = component->local_candidates; k; k = k->next) {
+        NiceCandidate *local_candidate = k->data;
+
+        if (nice_address_equal (&localaddr, &local_candidate->base_addr)) {
+          discovery_add_server_reflexive_candidate (
+              agent,
+              stream->id,
+              component->id,
+              &externaddr,
+              local_candidate->sockptr);
+          goto end;
+        }
+      }
+    }
+  }
+
+ end:
+  if (g_slist_length (agent->upnp_mapping)) {
+    if (agent->upnp_timer_source != NULL) {
+      g_source_destroy (agent->upnp_timer_source);
+      g_source_unref (agent->upnp_timer_source);
+      agent->upnp_timer_source = NULL;
+    }
+    agent_gathering_done (agent);
+  }
+
+  agent_unlock();
+}
+
+static void _upnp_error_mapping_port (GUPnPSimpleIgd *self, GError *error,
+    gchar *proto, guint external_port, gchar *local_ip, guint local_port,
+    gchar *description, gpointer user_data)
+{
+  NiceAgent *agent = (NiceAgent*)user_data;
+  NiceAddress localaddr;
+  GSList *i;
+
+  agent_lock();
+
+  nice_debug ("Agent %p : Error mapping %s:%d to %d (%d) : %s", agent, local_ip,
+      local_port, external_port, error->domain, error->message);
+  if (nice_address_set_from_string (&localaddr, local_ip)) {
+    nice_address_set_port (&localaddr, local_port);
+
+    for (i = agent->upnp_mapping; i; i = i->next) {
+      NiceAddress *addr = i->data;
+      if (nice_address_equal (&localaddr, addr)) {
+        agent->upnp_mapping = g_slist_remove (agent->upnp_mapping, addr);
+        nice_address_free (addr);
+        break;
+      }
+    }
+
+    if (g_slist_length (agent->upnp_mapping)) {
+      if (agent->upnp_timer_source != NULL) {
+        g_source_destroy (agent->upnp_timer_source);
+        g_source_unref (agent->upnp_timer_source);
+        agent->upnp_timer_source = NULL;
+      }
+      agent_gathering_done (agent);
+    }
+  }
+
+  agent_unlock();
+}
+
+#endif
 
 NICEAPI_EXPORT void
 nice_agent_gather_candidates (
@@ -1086,7 +1291,7 @@ nice_agent_gather_candidates (
   GSList *i;
   Stream *stream;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   stream = agent_find_stream (agent, stream_id);
   if (stream == NULL) {
@@ -1096,6 +1301,30 @@ nice_agent_gather_candidates (
   nice_debug ("Agent %p : In %s mode, starting candidate gathering.", agent,
       agent->full_mode ? "ICE-FULL" : "ICE-LITE");
 
+#ifdef HAVE_GUPNP
+  priv_free_upnp (agent);
+
+  if (agent->upnp_enabled) {
+    agent->upnp = gupnp_simple_igd_thread_new ();
+
+    agent->upnp_timer_source = agent_timeout_add_with_context (agent,
+        agent->upnp_timeout, priv_upnp_timeout_cb, agent);
+
+    if (agent->upnp) {
+      g_signal_connect (agent->upnp, "mapped-external-port",
+          G_CALLBACK (_upnp_mapped_external_port), agent);
+      g_signal_connect (agent->upnp, "error-mapping-port",
+          G_CALLBACK (_upnp_error_mapping_port), agent);
+    } else {
+      nice_debug ("Agent %p : Error creating UPnP Simple IGD agent", agent);
+    }
+  } else {
+    nice_debug ("Agent %p : UPnP property Disabled", agent);
+  }
+#else
+  nice_debug ("Agent %p : libnice compiled without UPnP support", agent);
+#endif
+
   /* if no local addresses added, generate them ourselves */
   if (agent->local_addresses == NULL) {
     GList *addresses = nice_interfaces_get_local_ips (FALSE);
@@ -1119,6 +1348,11 @@ nice_agent_gather_candidates (
     NiceAddress *addr = i->data;
     NiceCandidate *host_candidate;
 
+#ifdef HAVE_GUPNP
+    gchar local_ip[NICE_ADDRESS_STRING_LEN];
+    nice_address_to_string (addr, local_ip);
+#endif
+
     for (n = 0; n < stream->n_components; n++) {
       Component *component = stream_find_component_by_id (stream, n + 1);
       host_candidate = discovery_add_local_host_candidate (agent, stream->id,
@@ -1129,11 +1363,23 @@ nice_agent_gather_candidates (
         break;
       }
 
+#ifdef HAVE_GUPNP
+      if (agent->upnp_enabled) {
+        NiceAddress *addr = nice_address_dup (&host_candidate->base_addr);
+        nice_debug ("Agent %p: Adding UPnP port %s:%d", agent, local_ip,
+            nice_address_get_port (&host_candidate->base_addr));
+        gupnp_simple_igd_add_port (GUPNP_SIMPLE_IGD (agent->upnp), "UDP",
+            0, local_ip, nice_address_get_port (&host_candidate->base_addr),
+            0, PACKAGE_STRING);
+        agent->upnp_mapping = g_slist_prepend (agent->upnp_mapping, addr);
+      }
+#endif
+
       if (agent->full_mode &&
           agent->stun_server_ip) {
         NiceAddress stun_server;
         if (nice_address_set_from_string (&stun_server, agent->stun_server_ip)) {
-		  gboolean res;
+          gboolean res;
           nice_address_set_port (&stun_server, agent->stun_server_port);
 
           res =
@@ -1187,7 +1433,32 @@ nice_agent_gather_candidates (
 
  done:
 
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
+}
+
+static void priv_free_upnp (NiceAgent *agent)
+{
+#ifdef HAVE_GUPNP
+  GSList *i;
+
+  if (agent->upnp) {
+    g_object_unref (agent->upnp);
+    agent->upnp = NULL;
+  }
+
+  for (i = agent->upnp_mapping; i; i = i->next) {
+    NiceAddress *a = i->data;
+    nice_address_free (a);
+  }
+  g_slist_free (agent->upnp_mapping);
+  agent->upnp_mapping = NULL;
+
+  if (agent->upnp_timer_source != NULL) {
+    g_source_destroy (agent->upnp_timer_source);
+    g_source_unref (agent->upnp_timer_source);
+    agent->upnp_timer_source = NULL;
+  }
+#endif
 }
 
 static void priv_remove_keepalive_timer (NiceAgent *agent)
@@ -1209,7 +1480,7 @@ nice_agent_remove_stream (
   Stream *stream;
   GSList *i;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
   stream = agent_find_stream (agent, stream_id);
 
   if (!stream) {
@@ -1233,7 +1504,7 @@ nice_agent_remove_stream (
     priv_remove_keepalive_timer (agent);
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
 }
 
 NICEAPI_EXPORT gboolean
@@ -1243,7 +1514,7 @@ nice_agent_add_local_address (NiceAgent *agent, NiceAddress *addr)
   GSList *modified_list;
   gboolean ret = FALSE;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   dup = nice_address_dup (addr);
   nice_address_set_port (dup, 0);
@@ -1256,7 +1527,7 @@ nice_agent_add_local_address (NiceAgent *agent, NiceAddress *addr)
   }
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1293,6 +1564,22 @@ static gboolean priv_add_remote_candidate (
       strncpy(candidate->foundation, foundation, NICE_CANDIDATE_MAX_FOUNDATION);
     /* note: username and password must remain the same during
      *       a session; see sect 9.1.2 in ICE ID-19 */
+
+    /* note: however, the user/pass in ID-19 is global, if the user/pass
+     * are set in the candidate here, it means they need to be updated...
+     * this is essential to overcome a race condition where we might receive
+     * a valid binding request from a valid candidate that wasn't yet added to
+     * our list of candidates.. this 'update' will make the peer-rflx a
+     * server-rflx/host candidate again and restore that user/pass it needed
+     * to have in the first place */
+    if (username) {
+      g_free (candidate->username);
+      candidate->username = g_strdup (username);
+    }
+    if (password) {
+      g_free (candidate->password);
+      candidate->password = g_strdup (password);
+    }
     if (conn_check_add_for_candidate (agent, stream_id, component, candidate) < 0)
       error_flag = TRUE;
   }
@@ -1359,7 +1646,7 @@ nice_agent_set_remote_credentials (
   Stream *stream;
   gboolean ret = FALSE;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   stream = agent_find_stream (agent, stream_id);
   /* note: oddly enough, ufrag and pwd can be empty strings */
@@ -1373,7 +1660,7 @@ nice_agent_set_remote_credentials (
   }
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1387,7 +1674,7 @@ nice_agent_get_local_credentials (
   Stream *stream;
   gboolean ret = TRUE;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   stream = agent_find_stream (agent, stream_id);
   if (stream == NULL) {
@@ -1404,7 +1691,7 @@ nice_agent_get_local_credentials (
 
  done:
 
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1417,7 +1704,7 @@ nice_agent_set_remote_candidates (NiceAgent *agent, guint stream_id, guint compo
 
   nice_debug ("Agent %p: set_remote_candidates %d %d", agent, stream_id, component_id);
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   stream = agent_find_stream (agent, stream_id);
   if (stream == NULL) {
@@ -1461,7 +1748,7 @@ nice_agent_set_remote_candidates (NiceAgent *agent, guint stream_id, guint compo
  }
 
  done:
- g_static_rec_mutex_unlock (&agent->mutex);
+ agent_unlock();
  return added;
 }
 
@@ -1522,6 +1809,8 @@ _nice_agent_recv (
     }
   }
 
+  agent->media_after_tick = TRUE;
+
   if (stun_message_validate_buffer_length ((uint8_t *) buf, (size_t) len) != len)
     /* If the retval is no 0, its not a valid stun packet, probably data */
     return len;
@@ -1549,7 +1838,7 @@ nice_agent_send (
   Component *component;
   guint ret = -1;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   if (!agent_find_component (agent, stream_id, component_id, &stream, &component)) {
     goto done;
@@ -1578,7 +1867,7 @@ nice_agent_send (
     }
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1593,17 +1882,16 @@ nice_agent_get_local_candidates (
   GSList * ret = NULL;
   GSList * item = NULL;
 
-  g_static_rec_mutex_lock (&agent->mutex);
-  if (!agent_find_component (agent, stream_id, component_id, NULL, &component))
-    {
-      goto done;
-    }
+  agent_lock();
+  if (!agent_find_component (agent, stream_id, component_id, NULL, &component)) {
+    goto done;
+  }
 
   for (item = component->local_candidates; item; item = item->next)
     ret = g_slist_append (ret, nice_candidate_copy (item->data));
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1617,7 +1905,7 @@ nice_agent_get_remote_candidates (
   Component *component;
   GSList *ret = NULL, *item = NULL;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
   if (!agent_find_component (agent, stream_id, component_id, NULL, &component))
     {
       goto done;
@@ -1627,7 +1915,7 @@ nice_agent_get_remote_candidates (
     ret = g_slist_append (ret, nice_candidate_copy (item->data));
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1639,7 +1927,7 @@ nice_agent_restart (
   GSList *i;
   gboolean res = TRUE;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   /* step: clean up all connectivity checks */
   conn_check_free (agent);
@@ -1655,10 +1943,11 @@ nice_agent_restart (
     res = stream_restart (stream, agent->rng);
   }
 
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return res;
 }
 
+
 static void
 nice_agent_dispose (GObject *object)
 {
@@ -1702,10 +1991,11 @@ nice_agent_dispose (GObject *object)
   nice_rng_free (agent->rng);
   agent->rng = NULL;
 
+  priv_free_upnp (agent);
+
   if (G_OBJECT_CLASS (nice_agent_parent_class)->dispose)
     G_OBJECT_CLASS (nice_agent_parent_class)->dispose (object);
 
-  g_static_rec_mutex_free (&agent->mutex);
 }
 
 
@@ -1769,7 +2059,12 @@ nice_agent_g_source_cb (
   gchar buf[MAX_BUFFER_SIZE];
   gint len;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
+
+  if (g_source_is_destroyed (g_main_current_source ())) {
+    agent_unlock ();
+    return FALSE;
+  }
 
   /* note: dear compiler, these are for you: */
   (void)io;
@@ -1778,8 +2073,14 @@ nice_agent_g_source_cb (
 			  MAX_BUFFER_SIZE, buf);
 
   if (len > 0 && component->g_source_io_cb) {
-    component->g_source_io_cb (agent, stream->id, component->id,
-        len, buf, component->data);
+    gpointer data = component->data;
+    gint sid = stream->id;
+    gint cid = component->id;
+    NiceAgentRecvFunc callback = component->g_source_io_cb;
+    /* Unlock the agent before calling the callback */
+    agent_unlock();
+    callback (agent, sid, cid, len, buf, data);
+    goto done;
   } else if (len < 0) {
     GSource *source = ctx->source;
     component->gsources = g_slist_remove (component->gsources, source);
@@ -1789,9 +2090,13 @@ nice_agent_g_source_cb (
      * take care of every path where the socket might still be used.. */
     nice_debug ("Agent %p: unable to recv from socket %p. Detaching", agent,
         ctx->socket);
+
   }
 
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
+
+ done:
+
   return TRUE;
 }
 
@@ -1876,7 +2181,7 @@ nice_agent_attach_recv (
   Stream *stream = NULL;
   gboolean ret = FALSE;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   /* attach candidates */
 
@@ -1905,7 +2210,7 @@ nice_agent_attach_recv (
 
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1922,7 +2227,7 @@ nice_agent_set_selected_pair (
   CandidatePair pair;
   gboolean ret = FALSE;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   /* step: check that params specify an existing pair */
   if (!agent_find_component (agent, stream_id, component_id, &stream, &component)) {
@@ -1946,7 +2251,7 @@ nice_agent_set_selected_pair (
   ret = TRUE;
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
 
@@ -1980,7 +2285,7 @@ nice_agent_set_selected_remote_candidate (
   NiceCandidate *lcandidate = NULL;
   gboolean ret = FALSE;
 
-  g_static_rec_mutex_lock (&agent->mutex);
+  agent_lock();
 
   /* step: check if the component exists*/
   if (!agent_find_component (agent, stream_id, component_id, &stream, &component)) {
@@ -2007,6 +2312,6 @@ nice_agent_set_selected_remote_candidate (
   ret = TRUE;
 
  done:
-  g_static_rec_mutex_unlock (&agent->mutex);
+  agent_unlock();
   return ret;
 }
