@@ -495,7 +495,8 @@ gchar * nice_interfaces_get_ip_for_interface (gchar *interface_name)
           if_table->table[i].wszName, MAX_INTERFACE_NAME_LEN,
           NULL, NULL, NULL);
 
-      if (g_strcasecmp (interface_name, tmp_str) == 0) {
+      if (strlen (interface_name) == strlen (tmp_str) &&
+          g_ascii_strncasecmp (interface_name, tmp_str, strlen (interface_name)) == 0) {
         ret = win32_get_ip_for_interface (if_table->table[i].dwIndex);
         g_free (tmp_str);
         break;
